@@ -5,7 +5,7 @@ require 'tmpdir'
 class TestTmpdir < Test::Unit::TestCase
   def test_tmpdir_modifiable
     tmpdir = Dir.tmpdir
-    assert_equal(false, tmpdir.frozen?)
+    assert_not_predicate(tmpdir, :frozen?)
     tmpdir_org = tmpdir.dup
     tmpdir << "foo"
     assert_equal(tmpdir_org, Dir.tmpdir)
@@ -15,21 +15,34 @@ class TestTmpdir < Test::Unit::TestCase
     skip "no meaning on this platform" if /mswin|mingw/ =~ RUBY_PLATFORM
     Dir.mktmpdir do |tmpdir|
       # ToDo: fix for parallel test
-      olddir, ENV["TMPDIR"] = ENV["TMPDIR"], tmpdir
+      envs = %w[TMPDIR TMP TEMP]
+      oldenv = envs.each_with_object({}) {|v, h| h[v] = ENV.delete(v)}
       begin
-        assert_equal(tmpdir, Dir.tmpdir)
-        File.chmod(0777, tmpdir)
-        assert_not_equal(tmpdir, Dir.tmpdir)
-        newdir = Dir.mktmpdir("d", tmpdir) do |dir|
-          assert_file.directory? dir
-          assert_equal(tmpdir, File.dirname(dir))
-          dir
+        envs.each do |e|
+          tmpdirx = File.join(tmpdir, e)
+          ENV[e] = tmpdirx
+          assert_not_equal(tmpdirx, assert_warn('') {Dir.tmpdir})
+          File.write(tmpdirx, "")
+          assert_not_equal(tmpdirx, assert_warn(/not a directory/) {Dir.tmpdir})
+          File.unlink(tmpdirx)
+          ENV[e] = tmpdir
+          assert_equal(tmpdir, Dir.tmpdir)
+          File.chmod(0555, tmpdir)
+          assert_not_equal(tmpdir, assert_warn(/not writable/) {Dir.tmpdir})
+          File.chmod(0777, tmpdir)
+          assert_not_equal(tmpdir, assert_warn(/world-writable/) {Dir.tmpdir})
+          newdir = Dir.mktmpdir("d", tmpdir) do |dir|
+            assert_file.directory? dir
+            assert_equal(tmpdir, File.dirname(dir))
+            dir
+          end
+          assert_file.not_exist?(newdir)
+          File.chmod(01777, tmpdir)
+          assert_equal(tmpdir, Dir.tmpdir)
+          ENV[e] = nil
         end
-        assert_file.not_exist?(newdir)
-        File.chmod(01777, tmpdir)
-        assert_equal(tmpdir, Dir.tmpdir)
       ensure
-        ENV["TMPDIR"] = olddir
+        ENV.update(oldenv)
       end
     end
   end
